package com.yonge.cooleshow.biz.dal.service.impl;

import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.yonge.cooleshow.biz.dal.dto.UserAccountRecordDto;
import com.yonge.cooleshow.biz.dal.dto.req.TotalReq;
import com.yonge.cooleshow.biz.dal.entity.UserAccountRecord;
import com.yonge.cooleshow.biz.dal.entity.UserOrder;
import com.yonge.cooleshow.biz.dal.enums.InOrOutEnum;
import com.yonge.cooleshow.biz.dal.service.UserOrderService;
import com.yonge.cooleshow.biz.dal.vo.res.OrderPayRes;
import com.yonge.cooleshow.common.enums.CacheNameEnum;
import com.yonge.cooleshow.biz.dal.enums.FrozenTypeEnum;
import com.yonge.cooleshow.biz.dal.service.UserAccountRecordService;
import com.yonge.cooleshow.biz.dal.vo.res.AccountTotal;
import com.yonge.cooleshow.common.entity.HttpResponseResult;
import com.yonge.toolset.base.exception.BizException;
import com.yonge.toolset.payment.util.DistributedLock;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import com.yonge.cooleshow.biz.dal.entity.UserAccount;
import com.yonge.cooleshow.biz.dal.vo.UserAccountVo;
import com.yonge.cooleshow.biz.dal.dto.search.UserAccountSearch;
import com.yonge.cooleshow.biz.dal.dao.UserAccountDao;
import com.yonge.cooleshow.biz.dal.service.UserAccountService;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.*;


@Service
public class UserAccountServiceImpl extends ServiceImpl<UserAccountDao, UserAccount> implements UserAccountService {
    private final static Logger log = LoggerFactory.getLogger(UserAccountServiceImpl.class);

    @Autowired
    private UserAccountRecordService userAccountRecordService;
    @Autowired
    private RedissonClient redissonClient;
    @Autowired
    private UserOrderService orderService;

    @Override
    public UserAccountVo detail(Long id) {
        UserAccountVo detail = baseMapper.detail(id);
        if (null == detail) {
            throw new BizException("用户没有开通账户");
        }
        return detail;
    }

    @Override
    public IPage<UserAccountVo> selectPage(IPage<UserAccountVo> page, UserAccountSearch query) {
        return page.setRecords(baseMapper.selectPage(page, query));
    }

    @Override
    public HttpResponseResult<UserAccountRecord> accountChange(UserAccountRecordDto accountRecordDto) {
        if (null == accountRecordDto.getUserId() || null == accountRecordDto.getTransAmount()
                || null == accountRecordDto.getInOrOut() || null == accountRecordDto.getBizType()) {
            return HttpResponseResult.failed("缺少参数");
        }

        try {
            HttpResponseResult<UserAccountRecord> res = DistributedLock.of(redissonClient)
                    .runIfLockToFunction(CacheNameEnum.LOCK_CHANGE_ACCOUNT.getRedisKey(accountRecordDto.getUserId())
                            , this::doAccountChange, accountRecordDto, 10L);
            if (null != res) {
                return res;
            } else {
                return HttpResponseResult.failed("账户变更失败");
            }
        } catch (BizException e) {
            return HttpResponseResult.failed(e.getMessage());
        } catch (Exception e) {
            e.printStackTrace();
            return HttpResponseResult.failed("付款失败");
        }
    }

    @Override
    public HttpResponseResult<AccountTotal> accountTotal(TotalReq totalReq) {
        List<AccountTotal> infoList = baseMapper.accountTotal(totalReq.getTimeType().getCode(), totalReq);

        BigDecimal practiceAmount = BigDecimal.ZERO;
        BigDecimal liveAmount = BigDecimal.ZERO;
        BigDecimal videoAmount = BigDecimal.ZERO;
        BigDecimal musicAmount = BigDecimal.ZERO;
        for (AccountTotal info : infoList) {
            info.setPracticeAmount(null == info.getPracticeAmount() ? BigDecimal.ZERO : info.getPracticeAmount());
            info.setLiveAmount(null == info.getLiveAmount() ? BigDecimal.ZERO : info.getLiveAmount());
            info.setVideoAmount(null == info.getVideoAmount() ? BigDecimal.ZERO : info.getVideoAmount());
            info.setMusicAmount(null == info.getMusicAmount() ? BigDecimal.ZERO : info.getMusicAmount());

            practiceAmount = practiceAmount.add(info.getPracticeAmount());
            liveAmount = liveAmount.add(info.getLiveAmount());
            videoAmount = videoAmount.add(info.getVideoAmount());
            musicAmount = musicAmount.add(info.getMusicAmount());
        }
        AccountTotal total = new AccountTotal();

        total.setTotalInAmount(practiceAmount.add(liveAmount).add(videoAmount).add(musicAmount));
        if (total.getTotalInAmount().doubleValue() > 0) {
            total.setPracticeAmount(practiceAmount);
            BigDecimal practiceRate = total.getPracticeAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP).multiply(new BigDecimal("100"));
            total.setPracticeRate(practiceRate);

            total.setLiveAmount(liveAmount);
            BigDecimal liveRate = total.getLiveAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP).multiply(new BigDecimal("100"));
            total.setLiveRate(liveRate);

            total.setVideoAmount(videoAmount);
            BigDecimal videoRate = total.getVideoAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP).multiply(new BigDecimal("100"));
            total.setVideoRate(videoRate);

            total.setMusicAmount(musicAmount);
            BigDecimal musicRate = total.getMusicAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP).multiply(new BigDecimal("100"));
            total.setMusicRate(musicRate);
        }
        total.setInfoList(infoList);
        return HttpResponseResult.succeed(total);
    }

    @Transactional(rollbackFor = Exception.class)
    HttpResponseResult<UserAccountRecord> doAccountChange(UserAccountRecordDto accountRecordDto) {
        //收入要校验金额
        if (InOrOutEnum.IN.equals(accountRecordDto.getInOrOut())) {
            UserOrder userOrder = orderService.getOne(Wrappers.<UserOrder>lambdaQuery()
                    .eq(UserOrder::getOrderNo, accountRecordDto.getOrderNo()));

            BigDecimal totalTransAmount = baseMapper.totalTransAmount(accountRecordDto);
            if (null == totalTransAmount) {
                totalTransAmount = BigDecimal.ZERO;
            }
            //金额校验失败，入账为冻结金额
            if (null == userOrder || totalTransAmount.add(accountRecordDto.getTransAmount()).compareTo(userOrder.getActualPrice()) > 0) {
                log.error("账户变更失败,param is " + JSONObject.toJSONString(accountRecordDto));
                accountRecordDto.setErrFlag(1);
                accountRecordDto.setErrMsg("账户变更异常，订单入账总金额大于购买金额，订单号：" + accountRecordDto.getOrderNo());
            }
        }
        if (accountRecordDto.getErrFlag() == 0) {
            if (FrozenTypeEnum.FROZEN.equals(accountRecordDto.getFrozenType())) {
                baseMapper.frozenChangeAccount(accountRecordDto.getUserId(), accountRecordDto.getTransAmount(), accountRecordDto.getInOrOut().getCode());
            } else if (FrozenTypeEnum.FROZEN_DEDUCT.equals(accountRecordDto.getFrozenType())) {
                baseMapper.frozenDeductChangeAccount(accountRecordDto.getUserId(), accountRecordDto.getTransAmount(), accountRecordDto.getInOrOut().getCode());
            } else if (FrozenTypeEnum.FROZEN_BACK.equals(accountRecordDto.getFrozenType())) {
                baseMapper.frozenBackChangeAccount(accountRecordDto.getUserId(), accountRecordDto.getTransAmount(), accountRecordDto.getInOrOut().getCode());
            } else {
                baseMapper.changeAccount(accountRecordDto.getUserId(), accountRecordDto.getTransAmount(), accountRecordDto.getInOrOut().getCode());
            }
        }

        UserAccountVo detail = detail(accountRecordDto.getUserId());
        if (detail.getAmountTotal().doubleValue() < 0
                || detail.getAmountUsable().doubleValue() < 0 || detail.getAmountFrozen().doubleValue() < 0) {
            throw new BizException("账户变更失败");
        }

        if (accountRecordDto.getSaveRecord()) {
            //插入账户变更记录
            accountRecordDto.setAccountId(accountRecordDto.getUserId());
            accountRecordDto.setAccountBalance(detail.getAmountUsable());
            userAccountRecordService.save(accountRecordDto);
        }
        return HttpResponseResult.succeed(accountRecordDto);
    }
}